//-----------------------------------------------------------------------------
// file: TheForce.cpp
// desc: sample shows setting up OpenGL, modelview transform, 
//       basic real-time audio
//
// author: Music 256 | Fall 2008
//         CCRMA, Stanford University
//-----------------------------------------------------------------------------
#include <GLUT/glut.h>
#include <stdlib.h>
#include <memory.h>
#include <math.h>
#include <stdio.h>
#include "RtAudio.h"
#include <iostream>
#include <string>
#include <vector>
using namespace std;

// define sample
#define SAMPLE double
#define MY_PIE 3.14159265358979
#define MY_SRATE 44100

SAMPLE g_freq = 880;
SAMPLE g_phase = 0;
SAMPLE g_inc = g_freq / MY_SRATE;


struct Laser
{
    GLfloat x;
    GLfloat y;
    
    Laser( GLfloat _x, GLfloat _y ) : x(_x), y(_y) { }
};

// vector of lasers
vector<Laser> g_lasers;
GLfloat g_fire = 0.0;


// audio callback
int callme( char * buffer, int numFrames, void * user_data )
{
    // fprintf( stderr, "." );
    SAMPLE * buffy = (SAMPLE *)buffer;
    
    for( int i = 0; i < numFrames; i++ )
    {
        buffy[2*i] = buffy[2*i+1] = g_fire * sin( 2 * MY_PIE * g_phase );
        g_phase += g_inc;
        // quick hack for sound, use as envelope
        g_fire *= .9995;
        
        if( g_phase > 1.0f ) g_phase -= 1.0f;
    }
    
    return 0;    
}



//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void idleFunc( );
void displayFunc( );
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );

void initialize( );
void cleanUp( );




//-----------------------------------------------------------------------------
// global variables and #defines
//-----------------------------------------------------------------------------
#define __PI    3.1415926

// width and height of the window
GLsizei g_width = 800;
GLsizei g_height = 600;

// light 0 position
GLfloat g_light0_pos[4] = { 2.0f, 8.2f, 4.0f, 1.0f };

// light 1 parameters
GLfloat g_light1_ambient[] = { .2f, .2f, .2f, 1.0f };
GLfloat g_light1_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat g_light1_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat g_light1_pos[4] = { -2.0f, 0.0f, -4.0f, 1.0f };

// RtAudio object
RtAudio * g_audio;



//-----------------------------------------------------------------------------
// Name: main( )
// Desc: starting point
//-----------------------------------------------------------------------------
int main( int argc, char ** argv )
{
    // initialize GLUT
    glutInit( &argc, argv );
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "TheForce" );

    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );

    // do our own initialization
    initialize();

    // let GLUT handle the current thread from here
    glutMainLoop();

    return 0;
}




//-----------------------------------------------------------------------------
// Name: initialize( )
// Desc: sets initial OpenGL states
//       also initializes any application data
//-----------------------------------------------------------------------------
void initialize()
{
    // set the GL clear color - use when the color buffer is cleared
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
    glFrontFace( GL_CCW );
    // set fill mode
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    
    // enable lighting
    glEnable( GL_LIGHTING );
    // enable lighting for front
    glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
    // material have diffuse and ambient lighting 
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    // enable color
    glEnable( GL_COLOR_MATERIAL );
    
    // enable light 0
    glEnable( GL_LIGHT0 );
    
    // setup and enable light 1
    glLightfv( GL_LIGHT1, GL_AMBIENT, g_light1_ambient );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, g_light1_diffuse );
    glLightfv( GL_LIGHT1, GL_SPECULAR, g_light1_specular );
    // glEnable( GL_LIGHT1 );
    
    // set the position of the lights
    glLightfv( GL_LIGHT0, GL_POSITION, g_light0_pos );
    // glLightfv( GL_LIGHT1, GL_POSITION, g_light1_pos );
    
    // seed random number generator
    srand( time(NULL) );
    
    // buffer size
    int buffer_size = 512;
    
    // create the RtAudio
    try {
        g_audio = new RtAudio(
                            0, // device number of output
                            1, // number of output channels
                            0, // device number for input
                            0, // number of input channels
                            RTAUDIO_FLOAT64, // format
                            MY_SRATE, // sample rate
                            &buffer_size, // buffer size
                            8 // number of buffers
                            );
    } catch( RtError & err ) {
        err.printMessage();
        exit(1);
    }
    
    // set the callback
    try
    {
        g_audio->setStreamCallback( &callme, NULL );
        g_audio->startStream();
    }
    catch( RtError & err )
    {
        err.printMessage();
    }
}




//-----------------------------------------------------------------------------
// Name: cleanUp( )
// Desc: cleans up dynamic memory for vArray
//-----------------------------------------------------------------------------
void cleanUp( )
{
    // TODO: !!!
}





//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 0.0f, 10.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
}




//-----------------------------------------------------------------------------
// name: rand2f()
// desc: generate a random float
//-----------------------------------------------------------------------------
GLfloat rand2f( float a, float b )
{
    return a + (b-a)*(rand() / (GLfloat)RAND_MAX);
}




//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
    case 'Q':
    case 'q':
        exit(1);
    break;
    case 'f':
        // instantiate (local copy)
        Laser laser( 1.5, 0 );
        // copy/append
        g_lasers.push_back( laser );
        // set for audio
        g_fire = 1.0f;
    break;
    }

    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // when left mouse button is down, move left
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        // when right mouse button down, move right
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else
    {
    }

    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}




GLfloat g_rot = 0.0f;
GLfloat g_drot = 0.0f;
GLfloat g_mrot = 0.0f;
//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // push
    glPushMatrix();    
    // rotate everything
    glRotatef( g_rot, 0, 1, 0 ); g_rot += .05f;
    
    // draw lasers
    for( int i = 0; i < g_lasers.size(); i++ )
    {
        glPushMatrix();
        // position
        glTranslatef( g_lasers[i].x, g_lasers[i].y, 0 );
        // color it
        glColor3f( 1, .6, .6 );
        // draw it
        glutSolidCube( .25 );
        // move
        g_lasers[i].x -= .02;
        glPopMatrix();
    }

        // push
        glPushMatrix();
            // move left
            glTranslatef( 2, 0, 0 );
            // rotate it
            glRotatef( g_drot, 1, 0, 0 ); g_drot += .08;
            // draw deathstar
            glColor3f( .6, .6, .6 );
            glutSolidSphere( .5, 15, 15 );
            glTranslatef( -.501, 0, 0 );
            glRotatef( 90, 0, 1, 0 );
            glutSolidCone( .2, .01, 20, 20 );
        // pop
        glPopMatrix();
    
    // push
    glPushMatrix();
    // move left
    glTranslatef( -3, 0, 0 );
    
    // push
    glPushMatrix();
    // rotate moon
    glRotatef( g_mrot, 1, 0, 0 ); g_mrot += .04;
    // move the moon
    glTranslatef( 0, 2, 0 );
    // color the moon
    glColor3f( .25, .8, .25 );
    // draw moon
    glutSolidSphere( .25, 15, 15 );
    // pop
    glPopMatrix();
    
    // color alderan
    glColor3f( .25, .25, .8 );
    // rotate
    glRotatef( g_drot, 0, 0, 1 );
    // draw alderan
    glutSolidSphere( 1.5, 15, 15 );
    
    // pop
    glPopMatrix();
    
    // pop
    glPopMatrix();
    
    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
}





//     push
//    glPushMatrix();
//    // rotate
//    glRotatef( g_rotY, 0, 0, 1 );
//    // translate
//    glTranslatef( 3, 0, 0 );
//
//    // push
//    glPushMatrix();
//    // rotate
//    glRotatef( g_rotY*3, 0, 0, 1 );
//    // translate
//    glTranslatef( 1, 0, 0 );
//    // color it
//    glColor3f( .8, 1, .8 );
//    // draw
//    glutSolidTeapot( .2 );    
//    // pop
//    glPopMatrix();
//
//    // color it
//    glColor3f( .8, .8, 1 );
//    // draw
//    glutSolidTeapot( .2 );    
//    // pop
//    glPopMatrix();
//
//    // rotate
//    glRotatef( g_rotY, 0, 1, 0 );
//    // increment
//    g_rotY += .5;
//    // color it
//    glColor3f( 1, .8, .8 );
//    // draw sphere
//    glutSolidTeapot( 1 );